
from paymentsgate.transport import Response


class PaymentsgateError(Exception):
    pass


class APIError(PaymentsgateError):
    error: str
    message: str
    code: int
    data: object | None
    details: object | None
    
    def __init__(self, error: str, message: str, data: object | None, details: object | None, status: int) -> None: 
        super().__init__(f"[{error}] {message} (status: {status})")
        self.error = error
        self.message = message
        self.data = data
        self.code = status;
        self.details = details;

        if (details is not None):
            print('Error details:', self.details)
        if (data is not None):
            print(self.data)
        # print(f"{self.error}: {self.message} code: {self.code} details: {self.details}")
    


class APIResponseError(APIError):
    def __init__(self, response: Response) -> None:
        super().__init__(response.json.get('error'), response.json.get('message'), response.json.get('data'), response.json.get('details'),  response.status_code)

class APIAuthenticationError(APIError):
    def __init__(self, response: Response) -> None:
        super().__init__(response.json.get('error'), response.json.get('message'), response.json.get('data'), response.json.get('details'), response.status_code)



import logging
from paymentsgate.transport import Request, Response


def Logger(self, request: Request, response: Response):
    logging.debug(f"HTTP Request: {request}")
    logging.debug(f"HTTP Response: {response}")




from dataclasses import dataclass
from jwt import JWT
import time

@dataclass
class AccessToken:
    token: str
    expiredAt: int

    def __init__(self, token):
        self.token = token
        jwdInstance = JWT()
        parsed = jwdInstance.decode(token, do_verify=False, do_time_check=False)
        self.expiredAt = int(parsed['exp'])
    @property
    def is_expired(self):
        if self.expiredAt:
            return int(time.time()) >= self.expiredAt;
        return True

    def __str__(self) -> str:
        return self.token

@dataclass
class RefreshToken:
    token: str
    expiredAt: int

    def __init__(self, token, expiredAt):
        self.token = token
        self.expiredAt = expiredAt
    @property
    def is_expired(self):
        if self.expiredAt:
            return int(time.time()) >= self.expiredAt;
        return True

    def __str__(self) -> str:
        return self.token
    


from __future__ import annotations
from dataclasses import dataclass
import datetime
import json
from typing import Optional, List

from paymentsgate.enums import (
  Currencies, 
  InvoiceTypes, 
  Languages, 
  Statuses, 
  TTLUnits, 
  CurrencyTypes, 
  FeesStrategy,
  InvoiceDirection,
  CredentialsTypes
)

from pydantic import BaseModel, ConfigDict

@dataclass
class Credentials:
  def __init__(
      self, 
      account_id: str,
      merchant_id: str,
      project_id: str,
      private_key: str,
      public_key: str
  ):
     self.account_id = account_id
     self.merchant_id = merchant_id
     self.project_id = project_id
     self.private_key = private_key
     self.public_key = public_key
  
  @classmethod
  def fromFile(cls, filename):
    data = json.load(open(filename))
    return cls(data.get('account_id'),
               data.get('merchant_id'),
               data.get('project_id'),
               data.get('private_key'),
               data.get('public_key'))


@dataclass
class PayInFingerprintBrowserModel:
  acceptHeader: str
  colorDepth: int
  language: str
  screenHeight: int
  screenWidth: int
  timezone: str
  userAgent: str
  javaEnabled: bool
  windowHeight: int
  windowWidth: int

@dataclass
class PayInFingerprintModel:
  fingerprint: str
  ip: str
  country: str
  city: str
  state: str
  zip: str
  browser: Optional[PayInFingerprintBrowserModel]

@dataclass
class PayInModel:
  amount: float
  currency: Currencies
  invoiceId: Optional[str] # idempotent key
  clientId: Optional[str]
  type: InvoiceTypes
  bankId: Optional[str]
  trusted: Optional[bool]
  successUrl: Optional[str]
  failUrl: Optional[str]
  backUrl: Optional[str]
  clientCard: Optional[str]
  fingerprint: Optional[PayInFingerprintModel]
  lang: Optional[Languages]
   
@dataclass
class PayInResponseModel:
  id: str
  status: Statuses
  type: InvoiceTypes
  url: str

@dataclass
class PayOutRecipientModel:
  account_number: str
  account_owner: Optional[str]
  account_iban: Optional[str]
  account_swift: Optional[str]
  account_phone: Optional[str]
  account_bic: Optional[str]
  account_ewallet_name: Optional[str]
  account_email: Optional[str]
  account_bank_id: Optional[str]
  type: Optional[CredentialsTypes]

@dataclass
class PayOutModel:
  currency: Optional[Currencies] # currency from, by default = usdt
  currencyTo:Currencies
  amount: float
  invoiceId: Optional[str] # idempotent key
  clientId: Optional[str]
  ttl: Optional[int]
  ttl_unit: Optional[TTLUnits]
  finalAmount: Optional[float]
  sender_name: Optional[str]
  baseCurrency: Optional[CurrencyTypes]
  feesStrategy: Optional[FeesStrategy]
  recipient: PayOutRecipientModel
  quoteId: Optional[str]
   
@dataclass
class PayOutResponseModel:
  id: str
  status: Statuses

@dataclass
class GetQuoteModel:
  currency_from: Currencies
  currency_to: Currencies
  amount: float
  subtype: InvoiceTypes

@dataclass
class QuoteEntity:
  currency_from: CurrencyModel
  currency_to: CurrencyModel
  pair: str
  rate: float

@dataclass
class GetQuoteResponseModel:
  id: Optional[str]  = None
  finalAmount: Optional[float]  = None
  direction: Optional[InvoiceDirection]  = None
  fullRate: Optional[float]  = None
  fullRateReverse: Optional[float]  = None
  fees: Optional[float]  = None
  fees_percent: Optional[float]  = None
  quotes: Optional[List[QuoteEntity]]  = None
  expiredAt: Optional[datetime.datetime] = None

  #deprecated
  currency_from: Optional[CurrencyModel] = None
  currency_to: Optional[CurrencyModel] = None
  currency_middle: Optional[CurrencyModel] = None
  rate1: Optional[float] = None
  rate2: Optional[float] = None
  rate3: Optional[float] = None
  net_amount: Optional[float] = None
  metadata: Optional[object] = None

@dataclass
class DepositAddressResponseModel:
  currency: Currencies
  address: str
  expiredAt: datetime

@dataclass
class CurrencyModel:
  _id: str
  type: CurrencyTypes
  code: Currencies
  symbol: str
  label: Optional[str]
  decimal: int
  countryCode: Optional[str]
  countryName: Optional[str]

@dataclass
class BankModel:
  name: str
  title: str
  currency: Currencies
  fpsId: str

@dataclass
class InvoiceStatusModel:
  name: Statuses
  createdAt: datetime
  updatedAt: datetime

@dataclass
class InvoiceAmountModel:
  crypto: float
  fiat: float
  fiat_net: float

@dataclass
class InvoiceMetadataModel:
  invoiceId: Optional[str]
  clientId: Optional[str]
  fiatAmount: Optional[float]

@dataclass
class InvoiceModel:
  _id: str
  orderId: str
  projectId: str
  currencyFrom: CurrencyModel
  currencyTo: CurrencyModel
  direction: InvoiceDirection
  amount: float
  status: InvoiceStatusModel
  amounts: InvoiceAmountModel
  metadata: InvoiceMetadataModel
  receiptUrls: List[str]
  isExpired: bool
  createdAt: datetime
  updatedAt: datetime
  expiredAt: datetime

@dataclass
class AssetsAccountModel:
  currency: CurrencyModel;
  total: float
  pending: float
  available: float

@dataclass
class AssetsResponseModel:
  assets: List[AssetsAccountModel]




from dataclasses import dataclass
from pydantic import BaseModel

@dataclass
class Request:
    method: str
    path: str
    content_type: str = 'application/json'
    headers: dict[str, str] | None = None
    body: dict | None = None
    noAuth: bool | None = False
    signature: bool | None = False


@dataclass
class Response:
    raw_body: bytes
    json: dict
    status_code: int

    @property
    def success(self) -> bool:
        return self.status_code < 400

    def cast(self, model: BaseModel, error: dict):
        if self.success:
            return model(**self.json)
        return error(self.json.get('error'), self.json.get('message'), self.json.get('data'), self.json.get('status'));
    
    def __str__(self) -> str:
       return self.raw_body.decode("utf-8")


from __future__ import annotations
import logging
from dataclasses import dataclass, is_dataclass, field, asdict
import json
from urllib.parse import urlencode

from paymentsgate.tokens import (
  AccessToken,
  RefreshToken
)
from paymentsgate.exceptions import (
    APIResponseError, 
    APIAuthenticationError
)
from paymentsgate.models import (
    Credentials, 
    GetQuoteModel, 
    GetQuoteResponseModel, 
    PayInModel, 
    PayInResponseModel, 
    PayOutModel, 
    PayOutResponseModel,
    InvoiceModel
)
from paymentsgate.enums import ApiPaths
from paymentsgate.transport import (
  Request,
  Response
)
from paymentsgate.logger import Logger
from paymentsgate.cache import (
  AbstractCache, 
  DefaultCache
)

import requests

@dataclass
class ApiClient:
    baseUrl: str = field(default="", init=False)
    timeout: int = field(default=180, init=True)
    logger: Logger = Logger
    cache: AbstractCache = field(default_factory=DefaultCache)
    config: Credentials = field(default_factory=dict, init=False)

    REQUEST_DEBUG: bool = False
    RESPONSE_DEBUG: bool = False

    def __init__(self, config: Credentials, baseUrl: str, debug: bool=False):
        self.config = config
        self.cache = DefaultCache()
        self.baseUrl = baseUrl
        if debug:
            logging.basicConfig(level=logging.DEBUG)

    def PayIn(self, request: PayInModel) -> PayInResponseModel:
         # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.invoices_payin,
            content_type='application/json',
            noAuth=False,
            body=request,
        )

        # Handle response
        response = self._send_request(request)
        self.logger(request, response)
        if (response.success):
            return response.cast(PayInResponseModel, APIResponseError)
        else:
            raise APIResponseError(response)
        
    def PayOut(self, request: PayOutModel) -> PayOutResponseModel:
         # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.invoices_payout,
            content_type='application/json',
            noAuth=False,
            signature=True,
            body=request
        )

        # Handle response
        response = self._send_request(request)
        self.logger(request, response)
        if (response.success):
            return response.cast(PayOutResponseModel, APIResponseError)
        else:
            raise APIResponseError(response)

    def Quote(self, request: GetQuoteModel) -> GetQuoteResponseModel:
         # Prepare request
        request = Request(
            method="get",
            path=ApiPaths.fx_quote,
            content_type='application/json',
            noAuth=False,
            signature=False,
            body=request
        )

        # Handle response
        response = self._send_request(request)
        self.logger(request, response)
        if not response.success:
            raise APIResponseError(response)

        return response.cast(GetQuoteResponseModel, APIResponseError)
    
    def Status(self, query: str) -> InvoiceModel:
         # Prepare request
        request = Request(
            method="get",
            path=ApiPaths.invoices_info,
            content_type='application/json',
            noAuth=False,
            signature=False,
            body=request
        )

        # Handle response
        response = self._send_request(request)
        self.logger(request, response)
        if not response.success:
            raise APIResponseError(response)

        return response.cast(InvoiceModel, APIResponseError)

    @property
    def token(self) -> AccessToken | None:
        # First check if valid token is cached
        token = self.cache.get_token('access')
        refresh = self.cache.get_token('refresh')
        if token is not None and not token.is_expired:
            return token
        else:
            # try to refresh token
            if refresh is not None and not refresh.is_expired:
                refreshed = self._refresh_token()

                if (refreshed.success):
                    access = AccessToken(
                        response.json["access_token"]
                    )
                    refresh = RefreshToken(
                        response.json["refresh_token"],
                        int(response.json["expires_in"]),
                    )
                    self.cache.set_token(access)
                    self.cache.set_token(refresh)

                    return access

            # try to issue token            
            response = self._get_token()
            if response.success:
                
                access = AccessToken(
                    response.json["access_token"]
                )
                refresh = RefreshToken(
                    response.json["refresh_token"],
                    int(response.json["expires_in"]),
                )
                self.cache.set_token(access)
                self.cache.set_token(refresh)

                return access
            else:
                raise APIAuthenticationError(response)

    def _send_request(self, request: Request) -> Response:
        """
        Send a specified Request to the GoPay REST API and process the response
        """
        body = asdict(request.body) if is_dataclass(request.body) else request.body
        # Add Bearer authentication to headers if needed
        headers = request.headers or {}
        if not request.noAuth:
            auth = self.token
            if auth is not None:
                headers["Authorization"] = f"Bearer {auth.token}"

        if (request.method == 'get'):
            params = urlencode(body)
            r = requests.request(
                method=request.method,
                url=f"{self.baseUrl}{request.path}?{params}",
                headers=headers,
                timeout=self.timeout
            )
        else:
            r = requests.request(
                method=request.method,
                url=f"{self.baseUrl}{request.path}",
                headers=headers,
                json=body,
                timeout=self.timeout
            )

        # Build Response instance, try to decode body as JSON
        response = Response(raw_body=r.content, json={}, status_code=r.status_code)

        if (self.REQUEST_DEBUG):
            print(f"{request.method} => {self.baseUrl}{request.path} => {response.status_code}")
        
        try:
            response.json = r.json()
        except json.JSONDecodeError:
            pass

        self.logger(request, response)
        return response

    def _get_token(self) -> Response:
        # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.token_issue,
            content_type='application/json',
            noAuth=True,
            body={"account_id": self.config.account_id, "public_key": self.config.public_key},
        )
        # Handle response
        response = self._send_request(request)
        self.logger(request, response)
        return response
    
    def _refresh_token(self) -> Response:
         # Prepare request
        request = Request(
            method="post",
            path=ApiPaths.token_refresh,
            content_type='application/json',
            body={"refresh_token": self.refreshToken},
        )
        # Handle response
        response = self._send_request(request)
        self.logger(request, response)
        return response



from paymentsgate.client import ApiClient
from paymentsgate.enums import (
  AuthenticationRealms,
  ApiPaths,
  Currencies,
  Languages,
  Statuses,
  CurrencyTypes,
  InvoiceTypes,
  CredentialsTypes,
  RiskScoreLevels,
  CancellationReason,
  FeesStrategy,
  InvoiceDirection,
  TTLUnits
)
from paymentsgate.models import Credentials


from enum import Enum


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return self.value
    
class AuthenticationRealms(StrEnum):
    production = "production"
    sandbox = "sandbox"

class ApiPaths(StrEnum):
    token_issue = "/auth/token"
    token_refresh = "/auth/token/refresh"
    token_revoke = "/auth/token/revoke"
    token_validate = "/auth/token/validate"
    invoices_payin = "/deals/payin"
    invoices_payout = "/deals/payout"
    invoices_info = "/deals/:id"
    invoices_credentials = "/deals/:id/credentials"
    assets_list = "/wallet"
    assets_deposit = "/wallet/deposit"
    banks_list = "/banks/find"
    appel_create = "/support/create"
    appel_list = "/support/list"
    appel_stat = "/support/statistic"
    fx_quote = "/fx/calculatenew"

class Currencies(StrEnum):
    USDT = "USDT"
    EUR = "EUR"
    USD = "USD"
    TRY = "TRY"
    CNY = "CNY"
    JPY = "JPY"
    GEL = "GEL"
    AZN = "AZN"
    INR = "INR"
    AED = "AED"
    KZT = "KZT"
    UZS = "UZS"
    TJS = "TJS"
    EGP = "EGP"
    PKR = "PKR"
    IDR = "IDR"
    BDT = "BDT"
    GBP = "GBP"
    RUB = "RUB"
    THB = "THB"
    KGS = "KGS"
    PHP = "PHP"
    ZAR = "ZAR"
    ARS = "ARS"
    GHS = "GHS"
    KES = "KES"
    NGN = "NGN"
    AMD = "AMD"

class Languages(StrEnum):
    EN = "EN"
    IN = "IN"
    AE = "AE"
    TR = "TR"
    GE = "GE"
    RU = "RU"
    UZ = "UZ"
    AZ = "AZ"


class Statuses(StrEnum):
    queued = "queued"
    new = "new"
    pending = "pending"
    paid = "paid"
    completed = "completed"
    disputed = "disputed"
    canceled = "canceled"


class CurrencyTypes(StrEnum):
    fiat = "fiat"
    crypto = "crypto"


class InvoiceTypes(StrEnum):
    p2p = "p2p"
    ecom = "ecom"
    c2c = "c2c"
    m10 = "m10"
    mpay = "mpay"
    sbp = "sbp"
    sbpqr = "sbpqr"
    iban = "iban"
    upi = "upi"
    imps = "imps"
    spei = "spei"
    pix = "pix"
    rps = "rps"
    ibps = "ibps"
    bizum = "bizum"
    rkgs = "rkgs"
    kgsphone = "kgsphone"
    krungthainext = "krungthainext"
    sber = "sber"
    kztphone = "kztphone"
    accountbdt = "accountbdt"
    alipay = "alipay"
    accountegp = "accountegp"
    accountphp = "accountphp"
    sberqr = "sberqr"
    maya = "maya"
    gcash = "gcash"
    banktransferphp = "banktransferphp"
    banktransferars = "banktransferars"
    phonepe = "phonepe"
    freecharge = "freecharge"
    instapay = "instapay"
    vodafonecash = "vodafonecash"
    razn = "razn"
    rtjs = "rtjs"


class CredentialsTypes(StrEnum):
    iban = "iban"
    phone = "phone"
    card = "card"
    fps = "fps"
    account = "account"
    custom = "custom"


class RiskScoreLevels(StrEnum):
    unclassified = "unclassified"
    hr = "hr" # highest risk
    ftd = "ftd" # high risk
    trusted = "trusted" # low risk


class CancellationReason(StrEnum):
    NO_MONEY = "NO_MONEY"
    CREDENTIALS_INVALID = "CREDENTIALS_INVALID"
    EXPIRED = "EXPIRED"
    PRECHARGE_GAP_UPPER_LIMIT = "PRECHARGE_GAP_UPPER_LIMIT"
    CROSS_BANK_TFF_LESS_THAN_3K = "CROSS_BANK_TFF_LESS_THAN_3K"
    CROSS_BANK_UNSUPPORTED = "CROSS_BANK_UNSUPPORTED"


class FeesStrategy(StrEnum):
    add = "add"
    sub = "sub"


class InvoiceDirection(StrEnum):
    F2C = "F2C"
    C2F = "C2F"


class TTLUnits(StrEnum):
    sec = "sec"
    min = "min"
    hour = "hour"
